//--------------------------------------------------------------
// File     : stm32_ub_modul_s3m.c
// Datum    : 05.06.2014
// Version  : 1.0
// Autor    : UB
// EMail    : mc-4u(@)t-online.de
// Web      : www.mikrocontroller-4u.de
// CPU      : STM32F4
// IDE      : CooCox CoIDE 1.7.4
// GCC      : 4.7 2012q4
// Module   : keine
// Funktion : Modul-Typ : S3M
//--------------------------------------------------------------


//--------------------------------------------------------------
// Includes
//--------------------------------------------------------------
#include "stm32_ub_modul_s3m.h"


//-------------------------------------------------------------- 
static const uint16_t s3m_amigaPeriods[110] = {
 27392, 25856, 24384, 23040, 21696, 20480, 19328, 18240, 17216, 16256, 15360, 14496, // 0
 13696, 12928, 12192, 11520, 10848, 10240,  9664,  9120,  8608,  8128,  7680,  7248, // 1
  6848,  6464,  6096,  5760,  5424,  5120,  4832,  4560,  4304,  4064,  3840,  3624, // 2
  3424,  3232,  3048,  2880,  2712,  2560,  2416,  2280,  2152,  2032,  1920,  1812, // 3
  1712,  1616,  1524,  1440,  1356,  1280,  1208,  1140,  1076,  1016,   960,   906, // 4
   856,   808,   762,   720,   678,   640,   604,   570,   538,   508,   480,   453, // 5
   428,   404,   381,   360,   339,   320,   302,   285,   269,   254,   240,   226, // 6
   214,   202,   190,   180,   170,   160,   151,   143,   135,   127,   120,   113, // 7
   107,   101,    95,    90,    85,    80,    75,    71,    67,    63,    60,    56, // 8
     0,     0                                                                        // NONOTE, KEYOFF
};


//-------------------------------------------------------------- 
static const uint16_t s3m_fineTuneToHz[16] = {
 8363, 8413, 8463, 8529, 8581, 8651, 8723, 8757,
 7895, 7941, 7985, 8046, 8107, 8169, 8232, 8280
};


//-------------------------------------------------------------- 
static const uint8_t s3m_sine[64] = {
   0,  24,  49,  74,  97, 120, 141, 161,
 180, 197, 212, 224, 235, 244, 250, 253,
 255, 253, 250, 244, 235, 224, 212, 197,
 180, 161, 141, 120,  97,  74,  49,  24
};



//--------------------------------------------------------------
// interne Funktionen
//--------------------------------------------------------------
void s3m_initVar(void);
uint8_t s3m_loadHeader(UB_MOD_File *modfile);
void s3m_loadSamples(void);
void s3m_loadPattern(UB_MOD_File *modfile, uint8_t pattern);
void s3m_processRow(void);
void s3m_tremor(uint8_t channel);
void s3m_processTick(void);
void s3m_volumeSlide(uint8_t channel);
void s3m_portamento(uint8_t channel);
void s3m_vibrato(uint8_t channel, bool fine);
void s3m_tremolo(uint8_t channel);



//--------------------------------------------------------------
// init vom Modul-Typ : S3M
//
// ret_wert : 0 = ok
//            >0 = error
//--------------------------------------------------------------
uint8_t UB_S3M_Init(UB_MOD_File *modfile)
{
  uint8_t ret_wert=0;

  ret_wert=s3m_loadHeader(modfile);
  if(ret_wert!=0) {
    return(ret_wert);
  }
  s3m_loadSamples();
  s3m_initVar();

  return(ret_wert);
}


//--------------------------------------------------------------
// abspielen vom Modul-Typ : S3M
//
// loop : 0=without loop, 1=with loop
// ret_wert : 0 = playing , 1 = end of file
//--------------------------------------------------------------
uint8_t UB_S3M_Play(UB_MOD_File *modfile, uint8_t loop)
{

  if(S3m_player.tick == S3m_player.speed) {
    S3m_player.tick = 0;

    if(S3m_player.row == ROWS) {
      S3m_player.row = 0;
      S3m_player.orderIndex++;
      if(S3m_player.orderIndex == S3m.songLength) {
        S3m_player.orderIndex = 0;
	if(loop==0) return(1);   
      }
    }

    if(S3m_player.patternDelay) {
      S3m_player.patternDelay--;
    }
    else {
      if(S3m_player.orderIndex != S3m_player.oldOrderIndex) {
        s3m_loadPattern(modfile, S3m.order[S3m_player.orderIndex]);
      }
      S3m_player.oldOrderIndex = S3m_player.orderIndex;
      s3m_processRow();
    }
  }
  else {
    s3m_processTick();
  }

  S3m_player.tick++;

  return(0);
}


//--------------------------------------------------------------
// Mixer fuer Modul-Typ : S3M
//--------------------------------------------------------------
void UB_S3M_Mixer(UB_MOD_File *modfile)
{
  static int16_t sumL;
  static int16_t sumR;
  static uint8_t channel;
  static uint32_t samplePointer;
  static uint8_t current;
  static uint8_t next;
  static int16_t out;
  const uint8_t *ptr;

  sumL = 0;
  sumR = 0;
  for(channel = 0; channel < S3m.numberOfChannels; channel++) {
    if(!Mixer.channelFrequency[channel] ||
       !S3m.instruments[Mixer.channelSampleNumber[channel]].length) continue;

    Mixer.channelSampleOffset[channel] += Mixer.channelFrequency[channel];

    if(!Mixer.channelVolume[channel]) continue;

    samplePointer = Mixer.sampleBegin[Mixer.channelSampleNumber[channel]] +
             (Mixer.channelSampleOffset[channel] >> DIVIDER);

    if(Mixer.sampleLoopLength[Mixer.channelSampleNumber[channel]]) {
      if(samplePointer >= Mixer.sampleLoopEnd[Mixer.channelSampleNumber[channel]]) {
        Mixer.channelSampleOffset[channel] -= Mixer.sampleLoopLength[Mixer.channelSampleNumber[channel]] << DIVIDER;
        samplePointer -= Mixer.sampleLoopLength[Mixer.channelSampleNumber[channel]];
      }
    }
    else {
      if(samplePointer >= Mixer.sampleEnd[Mixer.channelSampleNumber[channel]]) {
        Mixer.channelFrequency[channel] = 0;
        samplePointer = Mixer.sampleEnd[Mixer.channelSampleNumber[channel]];
      }
    }

    ptr=modfile->table+samplePointer;

    current = *(ptr++);
    next = *(ptr++);

    // Unsigned to signed
    out = current-128;

    // Integer linear interpolation
    out += (next - current) * (Mixer.channelSampleOffset[channel] & ((1 << DIVIDER) - 1)) >> DIVIDER;

    // Upscale to BITDEPTH
    out <<= BITDEPTH - 8;

    // Channel volume
    out = out * Mixer.channelVolume[channel] >> 6;

    // Channel panning
    sumL += out * min(0xF - Mixer.channelPanning[channel], 8) >> 3;
    sumR += out * min(Mixer.channelPanning[channel], 8) >> 3;
  }

  // Downscale to BITDEPTH
  sumL>>=(BITDEPTH - 8);
  sumR>>=(BITDEPTH - 8);

  // Fill the sound buffer with unsigned values
  SoundBuffer.left[SoundBuffer.writePos] = sumL + (1 << (BITDEPTH - 1));
  SoundBuffer.right[SoundBuffer.writePos] = sumR + (1 << (BITDEPTH - 1));
  SoundBuffer.writePos++;
  SoundBuffer.writePos &= SOUNDBUFFERSIZE - 1;
}


//--------------------------------------------------------------
// interne Funktion
//--------------------------------------------------------------
void s3m_initVar(void)
{
  uint8_t channel;

  S3m_player.amiga = AMIGA_S3M;
  S3m_player.tick = S3m_player.speed;
  S3m_player.row = 0;

  S3m_player.orderIndex = 0;
  S3m_player.oldOrderIndex = 0xFF;
  S3m_player.patternDelay = 0;

  for(channel = 0; channel < S3m.numberOfChannels; channel++) {
    S3m_player.patternLoopCount[channel] = 0;
    S3m_player.patternLoopRow[channel] = 0;
    S3m_player.lastAmigaPeriod[channel] = 0;
    S3m_player.waveControl[channel] = 0;
    S3m_player.vibratoSpeed[channel] = 0;
    S3m_player.vibratoDepth[channel] = 0;
    S3m_player.vibratoPos[channel] = 0;
    S3m_player.tremoloSpeed[channel] = 0;
    S3m_player.tremoloDepth[channel] = 0;
    S3m_player.tremoloPos[channel] = 0;
    S3m_player.tremorOnOff[channel] = 0;
    S3m_player.tremorCount[channel] = 0;

    Mixer.channelSampleOffset[channel] = 0;
    Mixer.channelFrequency[channel] = 0;
    Mixer.channelVolume[channel] = 0;
  }

  SoundBuffer.writePos = 0;
  SoundBuffer.readPos = 0;
}


//--------------------------------------------------------------
// interne Funktion
// Header lesen
//
// ret_wert : 0 = ok
//            >0 = error
//--------------------------------------------------------------
uint8_t s3m_loadHeader(UB_MOD_File *modfile)
{
  uint8_t i;
  uint16_t songLength;
  uint16_t numberOfPatterns;
  uint16_t temp16;
  uint8_t masterVolume;
  uint8_t defaultPanning;
  uint8_t temp8;
  const uint8_t *ptr;
  uint8_t b0,b1;
  uint16_t n;

  if(modfile->size<1084) return(1);

  ptr=modfile->table; 

  for(n=0;n<S3M_NAME_LEN;n++) {
    S3m.name[n]=*(ptr++);
  }
  S3m.name[n]=0x00;

  // offset
  ptr=modfile->table+0x20;

  b0=*(ptr++);
  b1=*(ptr++);
  songLength = ((b1<<8)|b0);

  b0=*(ptr++);
  b1=*(ptr++);
  S3m.numberOfInstruments=((b1<<8)|b0);

  b0=*(ptr++);
  b1=*(ptr++);
  numberOfPatterns=((b1<<8)|b0);

  S3m.fastVolumeSlides = false;

  b0=*(ptr++);
  b1=*(ptr++);
  temp16=((b1<<8)|b0);
  if(temp16 & 64) S3m.fastVolumeSlides = true; 

  b0=*(ptr++);
  b1=*(ptr++);
  temp16=((b1<<8)|b0);
  if((temp16 & 0xFFF) == 300) S3m.fastVolumeSlides = true;

  // offset
  ptr=modfile->table+0x30;

  S3m.globalVolume=*(ptr++);
  S3m_player.speed=*(ptr++);
 
  temp8=*(ptr++);
  S3m_player.samplesPerTick = SAMPLERATE / (2 * temp8 / 5);

  masterVolume=*(ptr++);

  // dummy
  temp8=*(ptr++);

  defaultPanning=*(ptr++);

  // offset
  ptr=modfile->table+0x40;

  S3m.numberOfChannels = 0;
  memset(S3m.channelRemapping, 255, CHANNELS);
  for(i = 0; i < CHANNELS; i++) {
    temp8=*(ptr++);
    if(temp8 < 16) {
      S3m.channelRemapping[i] = S3m.numberOfChannels;
      if(temp8 < 8) {
        Mixer.channelPanning[S3m.numberOfChannels] = 0x3;
      }
      else {
        Mixer.channelPanning[S3m.numberOfChannels] = 0xC;
      }
      S3m.numberOfChannels++;
    }
  }
  if((S3m.numberOfChannels>CHANNELS) || (S3m.numberOfChannels==0)) return(3);

  // offset
  ptr=modfile->table+0x60;

  for(n=0;n<songLength;n++) {
    S3m.order[n]=*(ptr++);
  }

  S3m.songLength = 0;
  S3m.numberOfPatterns = 0;
  for(i = 0; i < songLength; i++) {
    S3m.order[S3m.songLength] = S3m.order[i];
    if(S3m.order[i] < 254) {
      S3m.songLength++;
      if(S3m.order[i] > S3m.numberOfPatterns) {
        S3m.numberOfPatterns = S3m.order[i];
      }
    }
  }
 
  for(n=0;n<S3m.numberOfInstruments;n++) {
    b0=*(ptr++);
    b1=*(ptr++);
    S3m.instrumentParapointers[n]=((b1<<8)|b0);
  }

  for(n=0;n<numberOfPatterns;n++) {
    b0=*(ptr++);
    b1=*(ptr++);
    S3m.patternParapointers[n]=((b1<<8)|b0);
  }

  if(defaultPanning == 0xFC) {
    for(i = 0; i < CHANNELS; i++) {
      temp8=*(ptr++);
      if((temp8 & 0x20) && S3m.channelRemapping[i]<CHANNELS) {
        Mixer.channelPanning[S3m.channelRemapping[i]] = temp8 & 0xF;
      }
    }
  }

  if(!(masterVolume & 128)) {
    for(i = 0; i < CHANNELS; i++) {
      Mixer.channelPanning[i] = 0x8;
    }
  }

  for(i = 0; i < INSTRUMENTS; i++) {
    S3m.instruments[i].middleC = 8363;
  }

  for(i = 0; i < S3m.numberOfInstruments; i++) {
    // offset
    ptr=modfile->table+((S3m.instrumentParapointers[i] << 4) + 13);

    // dummy
    temp8=*(ptr++);

    b0=*(ptr++);
    b1=*(ptr++);
    temp16=((b1<<8)|b0);
    S3m.instruments[i].sampleParapointer = temp16;

    b0=*(ptr++);
    b1=*(ptr++);
    S3m.instruments[i].length=((b1<<8)|b0);

    // dummy
    b0=*(ptr++);
    b1=*(ptr++);
    temp16=((b1<<8)|b0);

    b0=*(ptr++);
    b1=*(ptr++);
    S3m.instruments[i].loopBegin=((b1<<8)|b0);

    // dummy
    b0=*(ptr++);
    b1=*(ptr++);
    temp16=((b1<<8)|b0);

    b0=*(ptr++);
    b1=*(ptr++);
    S3m.instruments[i].loopEnd=((b1<<8)|b0);

    // dummy
    b0=*(ptr++);
    b1=*(ptr++);
    temp16=((b1<<8)|b0);

    S3m.instruments[i].volume=*(ptr++);

    // dummy
    b0=*(ptr++);
    b1=*(ptr++);
    temp16=((b1<<8)|b0);

    temp8=*(ptr++);
    S3m.instruments[i].loop = temp8 & 1;

    b0=*(ptr++);
    b1=*(ptr++);
    S3m.instruments[i].middleC=((b1<<8)|b0);

    if(!S3m.instruments[i].middleC) S3m.instruments[i].middleC = 8363; // Avoid division by zero

    // offset
    ptr+=14;

    for(n=0;n<S3M_NAME_LEN;n++) {
      S3m.instruments[i].name[n]=*(ptr++);
    }
    S3m.instruments[i].name[n]=0x00;

    if(S3m.instruments[i].loopEnd > S3m.instruments[i].length) {
      S3m.instruments[i].loopEnd = S3m.instruments[i].length;
    }
  }

  return(0);
}


//--------------------------------------------------------------
// interne Funktion
//--------------------------------------------------------------
void s3m_loadSamples(void)
{
  uint8_t i;
  uint32_t fileOffset;

  for(i = 0; i < S3m.numberOfInstruments; i++) {
    fileOffset = (S3m.instruments[i].sampleParapointer << 4) - 1;

    if(S3m.instruments[i].length) {
      Mixer.sampleBegin[i] = fileOffset;
      Mixer.sampleEnd[i] = fileOffset + S3m.instruments[i].length;
      if(S3m.instruments[i].loop && S3m.instruments[i].loopEnd - S3m.instruments[i].loopBegin > 2) {
        Mixer.sampleLoopBegin[i] = fileOffset + S3m.instruments[i].loopBegin;
        Mixer.sampleLoopLength[i] = S3m.instruments[i].loopEnd - S3m.instruments[i].loopBegin;
        Mixer.sampleLoopEnd[i] = fileOffset + S3m.instruments[i].loopEnd;
      }
      else {
        Mixer.sampleLoopBegin[i] = 0;
        Mixer.sampleLoopLength[i] = 0;
        Mixer.sampleLoopEnd[i] = 0;
      }
    }
  }
}


//--------------------------------------------------------------
// interne Funktion
//--------------------------------------------------------------
void s3m_loadPattern(UB_MOD_File *modfile, uint8_t pattern)
{
  static uint8_t row;
  static uint8_t channel;
  static uint8_t temp8;
  static uint16_t temp16;
  const uint8_t *ptr;

  // offset
  ptr=modfile->table+((S3m.patternParapointers[pattern] << 4) + 2);

  for(row = 0; row < ROWS; row++) {
    for(channel = 0; channel < CHANNELS; channel++) {
      S3m_player.currentPattern.note[row][channel] = S3M_NONOTE;
    }
  }
  memset(S3m_player.currentPattern.instrumentNumber, 0, ROWS * CHANNELS);
  memset(S3m_player.currentPattern.volume, S3M_NOVOLUME, ROWS * CHANNELS);
  memset(S3m_player.currentPattern.effectNumber, 0, ROWS * CHANNELS);
  memset(S3m_player.currentPattern.effectParameter, 0, ROWS * CHANNELS);

  row = 0;
  while(row < ROWS) {
    temp8=*(ptr++);
    if(temp8) {
      channel = S3m.channelRemapping[temp8 & 31];
      if(channel==0XFF) {
        //R.K.
      }
      else {
        if(temp8 & 32) {
          temp16=*(ptr++);
          switch(temp16) {
            case 255:
              S3m_player.currentPattern.note[row][channel] = S3M_NONOTE;
            break;
            case 254:
              S3m_player.currentPattern.note[row][channel] = S3M_KEYOFF;
            break;
            default:
              S3m_player.currentPattern.note[row][channel] = (temp16 >> 4) * 12 + (temp16 & 0xF);
          }
          S3m_player.currentPattern.instrumentNumber[row][channel]=*(ptr++);
        }
        if(temp8 & 64) {
          S3m_player.currentPattern.volume[row][channel]=*(ptr++);
        }
        if(temp8 & 128) {
          S3m_player.currentPattern.effectNumber[row][channel]=*(ptr++);
          S3m_player.currentPattern.effectParameter[row][channel]=*(ptr++);
        }
      }
    }
    else {
      row++;
    }
  }
}


//--------------------------------------------------------------
// interne Funktion
//--------------------------------------------------------------
void s3m_processRow(void)
{
  static uint8_t jumpFlag;
  static uint8_t breakFlag;
  static uint8_t channel;
  static uint16_t note;
  static uint8_t instrumentNumber;
  static uint8_t volume;
  static uint8_t effectNumber;
  static uint8_t effectParameter;
  static uint8_t effectParameterX;
  static uint8_t effectParameterY;
  static uint16_t sampleOffset;
  static uint8_t retriggerSample;

  S3m_player.lastRow = S3m_player.row++;
  jumpFlag = false;
  breakFlag = false;

  for(channel = 0; channel < S3m.numberOfChannels; channel++) {
    note = S3m_player.currentPattern.note[S3m_player.lastRow][channel];
    instrumentNumber = S3m_player.currentPattern.instrumentNumber[S3m_player.lastRow][channel];
    volume = S3m_player.currentPattern.volume[S3m_player.lastRow][channel];
    effectNumber = S3m_player.currentPattern.effectNumber[S3m_player.lastRow][channel];
    effectParameter = S3m_player.currentPattern.effectParameter[S3m_player.lastRow][channel];
    effectParameterX = effectParameter >> 4;
    effectParameterY = effectParameter & 0xF;
    sampleOffset = 0;

    if(instrumentNumber) {
      S3m_player.lastInstrumentNumber[channel] = instrumentNumber - 1;
      if(!(effectParameter == 0x13 && effectParameterX == S3M_NOTEDELAY)) {
        S3m_player.volume[channel] = S3m.instruments[S3m_player.lastInstrumentNumber[channel]].volume;
      }
    }

    if(note < S3M_NONOTE) {
      S3m_player.lastNote[channel] = note;
      S3m_player.amigaPeriod[channel] = s3m_amigaPeriods[note] * 8363 /
           S3m.instruments[S3m_player.lastInstrumentNumber[channel]].middleC;

      if(effectNumber != S3M_TONEPORTAMENTO && effectNumber != S3M_PORTAMENTOVOLUMESLIDE) {
        S3m_player.lastAmigaPeriod[channel] = S3m_player.amigaPeriod[channel];
      }

      if(!(S3m_player.waveControl[channel] & 0x80)) S3m_player.vibratoPos[channel] = 0;
      if(!(S3m_player.waveControl[channel] & 0x08)) S3m_player.tremoloPos[channel] = 0;
      S3m_player.tremorCount[channel] = 0;
      retriggerSample = true;
    }
    else {
      retriggerSample = false;
    }

    if(volume <= 0x40) S3m_player.volume[channel] = volume;
    if(note == S3M_KEYOFF) S3m_player.volume[channel] = 0;

    switch(effectNumber) {
      case S3M_SETSPEED:
        S3m_player.speed = effectParameter;
      break;

      case S3M_JUMPTOORDER:		
        S3m_player.orderIndex = effectParameter;
        if(S3m_player.orderIndex >= S3m.songLength) S3m_player.orderIndex = 0;
        S3m_player.row = 0;
        jumpFlag = true;
      break;

      case S3M_BREAKPATTERNTOROW:		 
        S3m_player.row = effectParameterX * 10 + effectParameterY;
        if(S3m_player.row >= ROWS) S3m_player.row = 0;
        if(!jumpFlag && !breakFlag) {
          S3m_player.orderIndex++;
          if(S3m_player.orderIndex >= S3m.songLength) S3m_player.orderIndex = 0;
        }
        breakFlag = true;
      break;

      case S3M_VOLUMESLIDE:
        if(effectParameter) S3m_player.lastVolumeSlide[channel] = effectParameter;
        if((S3m_player.lastVolumeSlide[channel] & 0xF) == 0xF) {
          S3m_player.volume[channel] += S3m_player.lastVolumeSlide[channel] >> 4;
        }
        else if(S3m_player.lastVolumeSlide[channel] >> 4 == 0xF) {
          S3m_player.volume[channel] -= S3m_player.lastVolumeSlide[channel] & 0xF;
        }
        if(S3m.fastVolumeSlides) {
          if(!(S3m_player.lastVolumeSlide[channel] & 0xF)) {
            S3m_player.volume[channel] += S3m_player.lastVolumeSlide[channel] >> 4;
          }
          if(!(S3m_player.lastVolumeSlide[channel] >> 4)) {
	      S3m_player.volume[channel] -= S3m_player.lastVolumeSlide[channel] & 0xF;
          }
        }
        if(S3m_player.volume[channel] > 64) {
          S3m_player.volume[channel] = 64;
        }
	else if(S3m_player.volume[channel] < 0) {
          S3m_player.volume[channel] = 0;
        }
      break;

      case S3M_PORTAMENTODOWN:
        if(effectParameter) S3m_player.lastPortamento[channel] = effectParameter;
        if(S3m_player.lastPortamento[channel] >> 4 == 0xF) {
          S3m_player.lastAmigaPeriod[channel] += (S3m_player.lastPortamento[channel] & 0xF) << 2;
        }
        if(S3m_player.lastPortamento[channel] >> 4 == 0xE) {
          S3m_player.lastAmigaPeriod[channel] += S3m_player.lastPortamento[channel] & 0xF;
        }
      break;

      case S3M_PORTAMENTOUP:
        if(effectParameter) S3m_player.lastPortamento[channel] = effectParameter;
        if(S3m_player.lastPortamento[channel] >> 4 == 0xF) {
          S3m_player.lastAmigaPeriod[channel] -= (S3m_player.lastPortamento[channel] & 0xF) << 2;
        }
        if(S3m_player.lastPortamento[channel] >> 4 == 0xE) {
          S3m_player.lastAmigaPeriod[channel] -= S3m_player.lastPortamento[channel] & 0xF;
        }
      break;

      case S3M_TONEPORTAMENTO:
        if(effectParameter) S3m_player.portamentoSpeed[channel] = effectParameter;
        S3m_player.portamentoNote[channel] = S3m_player.amigaPeriod[channel];
        retriggerSample = false;
      break;

      case S3M_VIBRATO:
        if(effectParameterX) S3m_player.vibratoSpeed[channel] = effectParameterX;
        if(effectParameterY) S3m_player.vibratoDepth[channel] = effectParameterY;
      break;

      case S3M_TREMOR:
        if(effectParameter) S3m_player.tremorOnOff[channel] = effectParameter;
        s3m_tremor(channel);
      break;

      case S3M_VIBRATOVOLUMESLIDE:
        if(effectParameter) S3m_player.lastVolumeSlide[channel] = effectParameter;
      break;

      case S3M_PORTAMENTOVOLUMESLIDE:
        if(effectParameter) S3m_player.lastVolumeSlide[channel] = effectParameter;
        S3m_player.portamentoNote[channel] = S3m_player.amigaPeriod[channel];
        retriggerSample = false;
      break;

      case S3M_SETSAMPLEOFFSET:
        sampleOffset = effectParameter << 8;
        if(sampleOffset > S3m.instruments[S3m_player.lastInstrumentNumber[channel]].length) {
          sampleOffset = S3m.instruments[S3m_player.lastInstrumentNumber[channel]].length;
        }
      break;

      case S3M_RETRIGGERNOTEVOLUMESLIDE:
        if(effectParameter) {
          S3m_player.retriggerVolumeSlide[channel] = effectParameterX;
          S3m_player.retriggerSpeed[channel] = effectParameterY;
        }
      break;

      case S3M_TREMOLO:
        if(effectParameterX) S3m_player.tremoloSpeed[channel] = effectParameterX;
        if(effectParameterY) S3m_player.tremoloDepth[channel] = effectParameterY;
      break;

      case S3M_SPECIALEFFECT_13:
        switch(effectParameterX) {
          case S3M_SETFINETUNE:
            S3m.instruments[S3m_player.lastInstrumentNumber[channel]].middleC = s3m_fineTuneToHz[effectParameterY];
          break;

          case S3M_SETVIBRATOWAVEFORM:
            S3m_player.waveControl[channel] &= 0xF0;
            S3m_player.waveControl[channel] |= effectParameterY;
          break;

          case S3M_SETTREMOLOWAVEFORM:
            S3m_player.waveControl[channel] &= 0xF;
            S3m_player.waveControl[channel] |= effectParameterY << 4;
          break;

          case S3M_SETCHANNELPANNING:
            Mixer.channelPanning[channel] = effectParameterY;
          break;

          case S3M_STEREOCONTROL:
            if(effectParameterY > 7) {
              effectParameterY -= 8;
            }
            else {
              effectParameterY += 8;
            }
            Mixer.channelPanning[channel] = effectParameterY;
          break;

          case S3M_PATTERNLOOP:
            if(effectParameterY) {
              if(S3m_player.patternLoopCount[channel]) {
                S3m_player.patternLoopCount[channel]--;
              }
              else {
                S3m_player.patternLoopCount[channel] = effectParameterY;
              }
              if(S3m_player.patternLoopCount[channel]) {
                S3m_player.row = S3m_player.patternLoopRow[channel] - 1;
              }
            }
            else {
              S3m_player.patternLoopRow[channel] = S3m_player.row;
            }
          break;

          case S3M_NOTEDELAY:
            retriggerSample = false;
          break;

          case S3M_PATTERNDELAY:
            S3m_player.patternDelay = effectParameterY;
          break;
        }
      break;

      case S3M_SETTEMPO:
        S3m_player.samplesPerTick = SAMPLERATE / ((effectParameter<<1) / 5);
      break;

      case S3M_FINEVIBRATO:
        if(effectParameterX) S3m_player.vibratoSpeed[channel] = effectParameterX;
        if(effectParameterY) S3m_player.vibratoDepth[channel] = effectParameterY;
      break;

      case S3M_SETGLOBALVOLUME:
      break;
    }

    if(retriggerSample) {
      Mixer.channelSampleOffset[channel] = sampleOffset << DIVIDER;
    }

    if((retriggerSample!=0) || ((S3m_player.lastAmigaPeriod[channel]) &&
        (effectNumber != S3M_VIBRATO) && (effectNumber != S3M_VIBRATOVOLUMESLIDE) &&
        (!((effectNumber == S3M_SPECIALEFFECT_13) && (effectParameterX == S3M_NOTEDELAY))) &&
        (effectNumber != S3M_FINEVIBRATO))) {

      Mixer.channelFrequency[channel] = S3m_player.amiga / S3m_player.lastAmigaPeriod[channel];
    }

    if(instrumentNumber) {
      Mixer.channelSampleNumber[channel] = S3m_player.lastInstrumentNumber[channel];
    }

    if(effectNumber != S3M_TREMOR && effectNumber != S3M_TREMOLO) {
      Mixer.channelVolume[channel] = S3m_player.volume[channel];
    }
  }
}


//--------------------------------------------------------------
// interne Funktion
//--------------------------------------------------------------
void s3m_tremor(uint8_t channel)
{
  uint8_t on = (S3m_player.tremorOnOff[channel] >> 4) + 1;
  uint8_t off = (S3m_player.tremorOnOff[channel] & 0xF) + 1;

  S3m_player.tremorCount[channel] %= on + off;
  if(S3m_player.tremorCount[channel] < on) {
    Mixer.channelVolume[channel] = S3m_player.volume[channel];
  }
  else {
    Mixer.channelVolume[channel] = 0;
  }
  S3m_player.tremorCount[channel]++;
}


//--------------------------------------------------------------
// interne Funktion
//--------------------------------------------------------------
void s3m_processTick(void)
{
  static uint8_t channel;
  static uint16_t note;
  static uint8_t instrumentNumber;
  static uint8_t volume;
  static uint8_t effectNumber;
  static uint8_t effectParameter;
  static uint8_t effectParameterX;
  static uint8_t effectParameterY;

  for(channel = 0; channel < S3m.numberOfChannels; channel++) {
    if(S3m_player.lastAmigaPeriod[channel]) {

      note = S3m_player.currentPattern.note[S3m_player.lastRow][channel];
      instrumentNumber = S3m_player.currentPattern.instrumentNumber[S3m_player.lastRow][channel];
      volume = S3m_player.currentPattern.volume[S3m_player.lastRow][channel];
      effectNumber = S3m_player.currentPattern.effectNumber[S3m_player.lastRow][channel];
      effectParameter = S3m_player.currentPattern.effectParameter[S3m_player.lastRow][channel];
      effectParameterX = effectParameter >> 4;
      effectParameterY = effectParameter & 0xF;

      switch(effectNumber) {
        case S3M_VOLUMESLIDE:
          s3m_volumeSlide(channel);
        break;

        case S3M_PORTAMENTODOWN:
          if(S3m_player.lastPortamento[channel] < 0xE0) {
            S3m_player.lastAmigaPeriod[channel] += S3m_player.lastPortamento[channel] << 2;
          }
          Mixer.channelFrequency[channel] = S3m_player.amiga / S3m_player.lastAmigaPeriod[channel];
        break;

        case S3M_PORTAMENTOUP:
          if(S3m_player.lastPortamento[channel] < 0xE0) {
            S3m_player.lastAmigaPeriod[channel] -= S3m_player.lastPortamento[channel] << 2;
          }
          Mixer.channelFrequency[channel] = S3m_player.amiga / S3m_player.lastAmigaPeriod[channel];
        break;

        case S3M_TONEPORTAMENTO:
          s3m_portamento(channel);
        break;

        case S3M_VIBRATO:
          s3m_vibrato(channel, false); // Fine = false
        break;

        case S3M_TREMOR:
          s3m_tremor(channel);
        break;

        case S3M_ARPEGGIO:
          if(effectParameter) {
            switch(S3m_player.tick % 3) {
              case 0:
                Mixer.channelFrequency[channel] = S3m_player.amiga / S3m_player.lastAmigaPeriod[channel];
              break;
              case 1:
                Mixer.channelFrequency[channel] = S3m_player.amiga / s3m_amigaPeriods[S3m_player.lastNote[channel] + effectParameterX];
              break;
              case 2:
                Mixer.channelFrequency[channel] = S3m_player.amiga / s3m_amigaPeriods[S3m_player.lastNote[channel] + effectParameterY];
              break;
            }
          }
        break;

        case S3M_VIBRATOVOLUMESLIDE:
          s3m_vibrato(channel, false); // Fine = false
          s3m_volumeSlide(channel);
        break;

        case S3M_PORTAMENTOVOLUMESLIDE:
          s3m_portamento(channel);
          s3m_volumeSlide(channel);
        break;

        case S3M_RETRIGGERNOTEVOLUMESLIDE:
          if(!S3m_player.retriggerSpeed[channel]) break;
          if(!(S3m_player.tick % S3m_player.retriggerSpeed[channel])) {
            if(S3m_player.retriggerVolumeSlide[channel]) {
              switch(S3m_player.retriggerVolumeSlide[channel]) {
                case 1:
                  S3m_player.volume[channel]--;
                break;
                case 2:
                  S3m_player.volume[channel] -= 2;
                break;
                case 3:
                  S3m_player.volume[channel] -= 4;
                break;
                case 4:
                  S3m_player.volume[channel] -= 8;
                break;
                case 5:
                  S3m_player.volume[channel] -= 16;
                break;
                case 6:
                  S3m_player.volume[channel] *= 2 / 3;
                break;
                case 7:
                  S3m_player.volume[channel] >>= 1;
                break;
                case 9:
                  S3m_player.volume[channel]++;
                break;
                case 0xA:
                  S3m_player.volume[channel] += 2;
                break;
                case 0xB:
                  S3m_player.volume[channel] += 4;
                break;
                case 0xC:
                  S3m_player.volume[channel] += 8;
                break;
                case 0xD:
                  S3m_player.volume[channel] += 16;
                break;
                case 0xE:
                  S3m_player.volume[channel] *= 3 / 2;
                break;
                case 0xF:
                  S3m_player.volume[channel] <<= 1;
                break;
              }
              if(S3m_player.volume[channel] > 64) {
                S3m_player.volume[channel] = 64;
              }
              else if(S3m_player.volume[channel] < 0) {
                S3m_player.volume[channel] = 0;
              }
              Mixer.channelVolume[channel] = S3m_player.volume[channel];
            }
            Mixer.channelSampleOffset[channel] = 0;
          }
        break;

        case S3M_TREMOLO:
          s3m_tremolo(channel);
        break;

        case S3M_SPECIALEFFECT_13:
          switch(effectParameterX) {
            case S3M_NOTECUT:
              if(S3m_player.tick == effectParameterY) {
                Mixer.channelVolume[channel] = S3m_player.volume[channel] = 0;
              }
            break;

            case S3M_NOTEDELAY:
              if(S3m_player.tick == effectParameterY) {
                if(instrumentNumber) S3m_player.volume[channel] = S3m.instruments[S3m_player.lastInstrumentNumber[channel]].volume;
                if(volume <= 0x40) S3m_player.volume[channel] = volume;
                if(note < S3M_NONOTE) Mixer.channelSampleOffset[channel] = 0;
                Mixer.channelFrequency[channel] = S3m_player.amiga / S3m_player.lastAmigaPeriod[channel];
                Mixer.channelVolume[channel] = S3m_player.volume[channel];
              }
            break;
          }
        break;

        case S3M_FINEVIBRATO:
          s3m_vibrato(channel, true); // Fine = true
        break;
      }
    }
  }
}


//--------------------------------------------------------------
// interne Funktion
//--------------------------------------------------------------
void s3m_volumeSlide(uint8_t channel)
{
  if(!(S3m_player.lastVolumeSlide[channel] & 0xF)) {
    S3m_player.volume[channel] += S3m_player.lastVolumeSlide[channel] >> 4;
  }
  if(!(S3m_player.lastVolumeSlide[channel] >> 4)) {
    S3m_player.volume[channel] -= S3m_player.lastVolumeSlide[channel] & 0xF;
  }

  if(S3m_player.volume[channel] > 64) {
    S3m_player.volume[channel] = 64;
  }
  else if(S3m_player.volume[channel] < 0) {
    S3m_player.volume[channel] = 0;
  }
  Mixer.channelVolume[channel] = S3m_player.volume[channel];
}


//--------------------------------------------------------------
// interne Funktion
//--------------------------------------------------------------
void s3m_portamento(uint8_t channel)
{
  if(S3m_player.lastAmigaPeriod[channel] < S3m_player.portamentoNote[channel]) {
    S3m_player.lastAmigaPeriod[channel] += S3m_player.portamentoSpeed[channel] << 2;
    if(S3m_player.lastAmigaPeriod[channel] > S3m_player.portamentoNote[channel]) {
      S3m_player.lastAmigaPeriod[channel] = S3m_player.portamentoNote[channel];
    }
  }
  if(S3m_player.lastAmigaPeriod[channel] > S3m_player.portamentoNote[channel]) {
    S3m_player.lastAmigaPeriod[channel] -= S3m_player.portamentoSpeed[channel] << 2;
    if(S3m_player.lastAmigaPeriod[channel] < S3m_player.portamentoNote[channel]) {
      S3m_player.lastAmigaPeriod[channel] = S3m_player.portamentoNote[channel];
    }
  }
  Mixer.channelFrequency[channel] = S3m_player.amiga / S3m_player.lastAmigaPeriod[channel];
}


//--------------------------------------------------------------
// interne Funktion
//--------------------------------------------------------------
void s3m_vibrato(uint8_t channel, bool fine)
{
  uint16_t delta;
  uint16_t temp;

  temp = S3m_player.vibratoPos[channel] & 31;

  switch(S3m_player.waveControl[channel] & 3) {
    case 0:
      delta = s3m_sine[temp];
    break;
    case 1:
      temp <<= 3;
      if(S3m_player.vibratoPos[channel] < 0) {
        temp = 255 - temp;
      }
      delta = temp;
    break;
    case 2:
      delta = 255;
    break;
    case 3:
      delta = rand() & 255;
    break;
  }

  delta *= S3m_player.vibratoDepth[channel];
  if(fine) {
    delta >>= 7;
  }
  else {
    delta >>= 5;
  }

  if(S3m_player.vibratoPos[channel] >= 0) {
    Mixer.channelFrequency[channel] = S3m_player.amiga / (S3m_player.lastAmigaPeriod[channel] + delta);
  }
  else {
    Mixer.channelFrequency[channel] = S3m_player.amiga / (S3m_player.lastAmigaPeriod[channel] - delta);
  }

  S3m_player.vibratoPos[channel] += S3m_player.vibratoSpeed[channel];
  if(S3m_player.vibratoPos[channel] > 31) S3m_player.vibratoPos[channel] -= 64;
}


//--------------------------------------------------------------
// interne Funktion
//--------------------------------------------------------------
void s3m_tremolo(uint8_t channel)
{
  uint16_t delta;
  uint16_t temp;

  temp = S3m_player.tremoloPos[channel] & 31;

  switch(S3m_player.waveControl[channel] & 3) {
    case 0:
      delta = s3m_sine[temp];
    break;
    case 1:
      temp <<= 3;
      if(S3m_player.tremoloPos[channel] < 0) {
        temp = 255 - temp;
      }
      delta = temp;
    break;
    case 2:
      delta = 255;
    break;
    case 3:
      delta = rand() & 255;
    break;
  }

  delta *= S3m_player.tremoloDepth[channel];
  delta >>= 6;

  if(S3m_player.tremoloPos[channel] >= 0) {
    if(S3m_player.volume[channel] + delta > 64) delta = 64 - S3m_player.volume[channel];
    Mixer.channelVolume[channel] = S3m_player.volume[channel] + delta;
  }
  else {
    if(S3m_player.volume[channel] - delta < 0) delta = S3m_player.volume[channel];
    Mixer.channelVolume[channel] = S3m_player.volume[channel] - delta;
  }

  S3m_player.tremoloPos[channel] += S3m_player.tremoloSpeed[channel];
  if(S3m_player.tremoloPos[channel] > 31) S3m_player.tremoloPos[channel] -= 64;
}




